from enum import Enum


class ValidationMode(Enum):
    NESTED_CV = "NESTED_CROSS_VALIDATION"
    CROSS_VALIDATION = "CROSS_VALIDATION"
    GRID_SEARCH = "GRID_SEARCH"


class RunMode(Enum):
    TRAIN = "train"
    # RE_EVAL = "re_eval"
    WARM_START = "warm_start"



from typing import Any, List, Dict
from ml_gym.batching.batch import EvaluationBatchResult
from ml_gym.error_handling.exception import BatchStateError
from ml_gym.gym.stateful_components import StatefulComponent
import numpy as np


class EarlyStoppingIF(StatefulComponent):

    def is_stopping_criterion_fulfilled(self, evaluation_results: List[EvaluationBatchResult], current_epoch: int) -> bool:
        raise NotImplementedError


class EarlyStopping(EarlyStoppingIF):
    def __init__(self, strategy: EarlyStoppingIF):
        self.strategy = strategy

    def is_stopping_criterion_fulfilled(self, evaluation_results: List[EvaluationBatchResult], current_epoch: int) -> bool:
        return self.strategy.is_stopping_criterion_fulfilled(evaluation_results=evaluation_results, current_epoch=current_epoch)

    def get_state(self) -> Dict[str, Any]:
        state = self.strategy.get_state()
        return state

    def set_state(self, state: Dict[str, Any]):
        self.strategy.set_state(state)


class LastKEpochsImprovementStrategy(EarlyStoppingIF):

    def __init__(self, min_relative_improvement: float, epochs_window: int, split_name: str, monitoring_key: str,
                 is_increase_task: bool):
        monitoring_values_list = [1]
        init_min_relative_improvement = min_relative_improvement * 1.1  # we need to make this a bit bigger for precision inaccuracies
        for i in range(1, epochs_window):
            monitoring_values_list.append(monitoring_values_list[i-1] + monitoring_values_list[i-1]*init_min_relative_improvement)
        self.monitoring_values = np.array(monitoring_values_list)
        if not is_increase_task:
            self.monitoring_values = self.monitoring_values[::-1]
        self.min_relative_improvement = min_relative_improvement
        self.epochs_window = epochs_window
        self.split_name = split_name
        self.monitoring_key = monitoring_key
        self.is_increase_task = is_increase_task

    def _get_monitoring_value(self, evaluation_results: List[EvaluationBatchResult]) -> float:
        evaluation_result = None
        for e in evaluation_results:
            if e.split_name == self.split_name:
                evaluation_result = e
                break
        try:
            value = evaluation_result.metrics[self.monitoring_key][-1] if self.monitoring_key in evaluation_result.metrics else evaluation_result.losses[self.monitoring_key][-1]
        except KeyError as e:
            raise BatchStateError(f"Monitoring key {self.monitoring_key} not present in metrics or losses.") from e
        return value

    def _evaluate_history(self, current_epoch: int) -> bool:
        def monitoring_diff_fun(i, current_epoch):
            next_val = self.monitoring_values[(i + 2 + current_epoch) % self.epochs_window]
            base_val = self.monitoring_values[(i + 1 + current_epoch) % self.epochs_window]
            relative_diff = (next_val - base_val) / base_val
            return relative_diff

        relative_diffs = [monitoring_diff_fun(i, current_epoch) for i in range(self.epochs_window-1)]
        if self.is_increase_task:
            return max(relative_diffs) < self.min_relative_improvement
        else:
            return min(relative_diffs) > self.min_relative_improvement * -1

    def is_stopping_criterion_fulfilled(self, evaluation_results: List[EvaluationBatchResult], current_epoch: int) -> bool:
        monitoring_value = self._get_monitoring_value(evaluation_results)
        self.monitoring_values[current_epoch % self.epochs_window] = monitoring_value

        perform_stop = self._evaluate_history(current_epoch)
        return perform_stop

    def get_state(self) -> Dict[str, Any]:
        state = {"monitoring_values": self.monitoring_values.tolist()}
        return state

    def set_state(self, state: Dict[str, Any]):
        self.monitoring_values = np.array(state["monitoring_values"])


class EarlyStoppingStrategyFactory:

    def get_last_k_epochs_improvement_strategy(min_relative_improvement: float, epochs_window: int, split_name: str,
                                               monitoring_key: str, is_increase_task: bool) -> EarlyStoppingIF:
        strategy = LastKEpochsImprovementStrategy(min_relative_improvement=min_relative_improvement,
                                                  epochs_window=epochs_window,
                                                  split_name=split_name,
                                                  monitoring_key=monitoring_key,
                                                  is_increase_task=is_increase_task)
        return EarlyStopping(strategy=strategy)



from abc import ABC, abstractmethod
from ml_gym.gym.jobs import GymJob
from typing import List, Type, Dict, Any
from ml_gym.modes import RunMode
from ml_gym.persistency.logging import MLgymStatusLoggerCollectionConstructable
import torch
from ml_gym.persistency.io import GridSearchAPIClientConstructableIF


class BluePrint(ABC):
    """ Abstract class that provides a blueprint for creating `AbstractGymJob`
    """

    def __init__(self, run_mode: RunMode, num_epochs: int,
                 config: Dict[str, Any], grid_search_id: str,
                 gs_api_client_constructable: GridSearchAPIClientConstructableIF,
                 experiment_id: str,
                 external_injection: Dict[str, Any] = None,
                 logger_collection_constructable: MLgymStatusLoggerCollectionConstructable = None,
                 warm_start_epoch: int = 0):

        self.run_mode = run_mode
        self.config = config
        self.grid_search_id = grid_search_id
        self.experiment_id = experiment_id
        self.num_epochs = num_epochs
        self.external_injection = external_injection if external_injection is not None else {}
        self.logger_collection_constructable = logger_collection_constructable
        self.warm_start_epoch = warm_start_epoch
        self.gs_api_client_constructable = gs_api_client_constructable

    @abstractmethod
    def construct(self, device: torch.device = None) -> GymJob:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def construct_components(config: Dict, component_names: List[str], device: torch.device,
                             external_injection: Dict[str, Any] = None) -> List[Any]:
        return NotImplementedError

    @staticmethod
    def create_blueprint(blue_print_class: Type["BluePrint"],
                         run_mode: RunMode,
                         experiment_config: Dict[str, Any],
                         experiment_id: str,
                         num_epochs: int,
                         grid_search_id: str,
                         gs_api_client_constructable: GridSearchAPIClientConstructableIF,
                         external_injection: Dict[str, Any] = None,
                         logger_collection_constructable: MLgymStatusLoggerCollectionConstructable = None,
                         warm_start_epoch: int = 0) -> List["BluePrint"]:

        blue_print = blue_print_class(grid_search_id=grid_search_id,
                                      experiment_id=experiment_id,
                                      num_epochs=num_epochs,
                                      warm_start_epoch=warm_start_epoch,
                                      run_mode=run_mode,
                                      config=experiment_config,
                                      external_injection=external_injection,
                                      logger_collection_constructable=logger_collection_constructable,
                                      gs_api_client_constructable=gs_api_client_constructable)
        return blue_print



import json
from typing import List, Type
from ml_board.backend.restful_api.data_models import FileFormat
from ml_gym.blueprints.blue_prints import BluePrint
from ml_gym.gym.gym import Gym
from ml_gym.modes import RunMode
from ml_gym.persistency.io import GridSearchAPIClientConstructableIF, GridSearchAPIClientIF
from ml_gym.persistency.logging import LoggerConstructableIF, MLgymStatusLoggerCollectionConstructable
from ml_gym.util.logger import QueuedLogging
from multiprocessing import Queue
from ml_gym.validation.validator import ValidatorIF
from datetime import datetime
from ml_gym.io.config_parser import YAMLConfigLoader


class MLGymStarter:
    @staticmethod
    def _create_gym(job_id_prefix: str, process_count: int, device_ids, log_std_to_file: bool,
                    logger_collection_constructable: MLgymStatusLoggerCollectionConstructable) -> Gym:
        gym = Gym(job_id_prefix, process_count, device_ids=device_ids, log_std_to_file=log_std_to_file,
                  logger_collection_constructable=logger_collection_constructable)
        return gym

    @staticmethod
    def _setup_logging_environment(log_dir_path: str):
        if QueuedLogging._instance is None:
            queue = Queue()
            QueuedLogging.start_logging(queue, log_dir_path)

    @staticmethod
    def _stop_logging_environment():
        QueuedLogging.stop_listener()


# class MLGymTrainWarmStarter(MLGymStarter):

#     def __init__(self, grid_search_id: str, blue_print_class: Type[BluePrint], num_epochs: int, text_logging_path: str, process_count: int,
#                  gpus: List[int], log_std_to_file: bool, gs_api_client_constructable: GridSearchAPIClientConstructableIF,
#                  logger_collection_constructable: MLgymStatusLoggerCollectionConstructable = None) -> None:
#         self.blue_print_class = blue_print_class
#         self.num_epochs = num_epochs
#         self.text_logging_path = text_logging_path
#         self.process_count = process_count
#         self.log_std_to_file = log_std_to_file
#         self.gpus = gpus
#         self.grid_search_id = grid_search_id
#         self.gs_api_client_constructable = gs_api_client_constructable
#         self.gs_api_client = gs_api_client_constructable.construct()
#         self.logger_collection_constructable = logger_collection_constructable

#     def start(self):
#         self._setup_logging_environment(self.text_logging_path)
#         gym = MLGymStarter._create_gym(process_count=self.process_count, device_ids=self.gpus, log_std_to_file=self.log_std_to_file,
#                                        logger_collection_constructable=self.logger_collection_constructable)

#         blueprints = self.validator.create_blueprints(blue_print_type=self.blue_print_class,
#                                                       gs_config=self.gs_config,
#                                                       num_epochs=self.num_epochs,
#                                                       logger_collection_constructable=self.logger_collection_constructable)

#         gym.add_blue_prints(blueprints)
#         gym.run(parallel=True)

#         self._stop_logging_environment()


class MLGymTrainStarter(MLGymStarter):

    def __init__(self, text_logging_path: str, process_count: int,
                 gpus: List[int], log_std_to_file: bool, blueprints: List[BluePrint],
                 logger_collection_constructable: MLgymStatusLoggerCollectionConstructable = None) -> None:
        self.text_logging_path = text_logging_path
        self.process_count = process_count
        self.log_std_to_file = log_std_to_file
        self.gpus = gpus
        self.blueprints = blueprints
        self.logger_collection_constructable = logger_collection_constructable

    def start(self):
        self._setup_logging_environment(self.text_logging_path)
        job_id_prefix = datetime.now().strftime("%Y-%m-%d--%H-%M-%S")

        gym = MLGymStarter._create_gym(job_id_prefix=job_id_prefix, process_count=self.process_count, device_ids=self.gpus,
                                       log_std_to_file=self.log_std_to_file,
                                       logger_collection_constructable=self.logger_collection_constructable)
        gym.add_blueprints(self.blueprints)
        gym.run(parallel=True)

        self._stop_logging_environment()


def save_blueprint_config(blueprint: BluePrint, gs_api_client: GridSearchAPIClientIF):
    gs_api_client.add_config_string(grid_search_id=blueprint.grid_search_id, config_name="experiment_config.json",
                                    config=json.dumps(blueprint.config), experiment_id=blueprint.experiment_id,
                                    file_format=FileFormat.JSON)


def mlgym_entry_train(blueprint_class: Type[BluePrint],
                      logger_collection_constructable: LoggerConstructableIF,
                      gs_api_client_constructable: GridSearchAPIClientConstructableIF,
                      gs_config_raw_string: str,
                      validator: ValidatorIF,
                      text_logging_path: str,
                      process_count: int,
                      gpus: int,
                      log_std_to_file: bool,
                      num_epochs: int,
                      validation_strategy_config_raw_string: str = None):
    gs_api_client = gs_api_client_constructable.construct()

    grid_search_id = datetime.now().strftime("%Y-%m-%d--%H-%M-%S")
    gs_config = YAMLConfigLoader.load_string(gs_config_raw_string)
    gs_api_client.add_config_string(grid_search_id=grid_search_id,
                                    file_format=FileFormat.YAML,
                                    config_name="grid_search_config.yml",
                                    config=gs_config_raw_string)

    if validation_strategy_config_raw_string is not None:
        gs_api_client.add_config_string(grid_search_id=grid_search_id,
                                        config_name="validation_strategy_config.yml",
                                        config=validation_strategy_config_raw_string)

    blueprints = validator.create_blueprints(grid_search_id=grid_search_id,
                                             blue_print_type=blueprint_class,
                                             gs_config=gs_config,
                                             num_epochs=num_epochs,
                                             gs_api_client_constructable=gs_api_client_constructable,
                                             logger_collection_constructable=logger_collection_constructable)

    for blueprint in blueprints:
        save_blueprint_config(blueprint=blueprint, gs_api_client=gs_api_client)

    starter = MLGymTrainStarter(text_logging_path=text_logging_path,
                                process_count=process_count,
                                gpus=gpus,
                                log_std_to_file=log_std_to_file,
                                blueprints=blueprints,
                                logger_collection_constructable=logger_collection_constructable)
    starter.start()


def mlgym_entry_warm_start(blueprint_class: Type[BluePrint],
                           grid_search_id: str,
                           logger_collection_constructable: LoggerConstructableIF,
                           gs_api_client_constructable: GridSearchAPIClientConstructableIF,
                           text_logging_path: str,
                           process_count: int,
                           gpus: int,
                           log_std_to_file: bool,
                           num_epochs: int):
    gs_api_client = gs_api_client_constructable.construct()
    experiment_statuses = gs_api_client.get_experiment_statuses(grid_search_id)

    blueprints = [BluePrint.create_blueprint(blue_print_class=blueprint_class,
                                             run_mode=RunMode.WARM_START,
                                             experiment_config=experiment_status.experiment_config,
                                             num_epochs=num_epochs,
                                             warm_start_epoch=experiment_status.last_checkpoint_id,
                                             grid_search_id=grid_search_id,
                                             experiment_id=experiment_status.experiment_id,
                                             logger_collection_constructable=logger_collection_constructable,
                                             gs_api_client_constructable=gs_api_client_constructable)
                  for experiment_status in experiment_statuses if experiment_status.last_checkpoint_id < num_epochs]

    starter = MLGymTrainStarter(text_logging_path=text_logging_path,
                                process_count=process_count,
                                gpus=gpus,
                                log_std_to_file=log_std_to_file,
                                blueprints=blueprints,
                                logger_collection_constructable=logger_collection_constructable)
    starter.start()



from abc import abstractmethod
from typing import Dict, BinaryIO
from dashify.logging.dashify_logging import ExperimentInfo
from ml_gym.persistency.io import DashifyWriter


class StateTrackableIF:

    @abstractmethod
    def get_state_representation(self) -> BinaryIO:
        raise NotImplementedError

    @abstractmethod
    def load_state_representation(self, representation: BinaryIO):
        raise NotImplementedError


class StateTrackerIF:

    @abstractmethod
    def register_trackable(self, trackable: StateTrackableIF):
        raise NotImplementedError

    @abstractmethod
    def save_state(self, measurement_id: int = 0):
        raise NotImplementedError


class StateTracker(StateTrackerIF):

    def __init__(self):
        super().__init__()
        self.trackables: Dict[str, StateTrackableIF] = {}

    def register_trackable(self, key: str, trackable: StateTrackableIF):
        self.trackables[key] = trackable


class DiscStateTracker(StateTracker):

    def __init__(self, experiment_info: ExperimentInfo):
        super().__init__()
        self.experiment_info = experiment_info
        self.trackables: Dict[str, StateTrackableIF] = {}

    def save_state(self, measurement_id: int = 0):
        states = {key: trackable.get_state_representation() for key,  trackable in self.trackables.items()}
        for key, state in states.items():
            DashifyWriter.save_binary_state(key=key,
                                            state=state,
                                            experiment_info=self.experiment_info,
                                            measurement_id=measurement_id)



from torch.multiprocessing import Queue
import tqdm
import torch
from typing import List
from ml_gym.multiprocessing.job import JobType, Job, JobCollection, JobStatusSubscriberIF
from ml_gym.multiprocessing.worker import WorkerProcessWrapper
from ml_gym.util.logger import QueuedLogging
from ml_gym.util.logger import LogLevel, QLogger
from ml_gym.multiprocessing.job import JobStatus
from ml_gym.persistency.logging import JobStatusLogger, MLgymStatusLoggerCollectionConstructable


class JobStatusLoggingSubscriber(JobStatusSubscriberIF):

    def __init__(self, logger: JobStatusLogger):
        self._logger = logger

    def callback_job_event(self, job: Job):
        representation = {"job_id": job.job_id, "job_type": job.job_type, "grid_search_id": job.grid_search_id, "experiment_id": job.experiment_id, "status": job.status,
                          "starting_time": job.starting_time, "finishing_time": job.finishing_time, "error": job.error,
                          "stacktrace": job.stacktrace, "device": job.device}
        self._logger.log_job_status(**representation)


class Pool:
    def __init__(self, num_processes: int, devices: List[torch.device], max_jobs_per_process: int = 1,
                 logger_collection_constructable: MLgymStatusLoggerCollectionConstructable = None):
        self.num_processes = num_processes
        self.job_q = Queue()
        self.job_update_q = Queue()
        self.devices = devices
        self.worker_processes = []
        self.max_jobs_per_process = max_jobs_per_process
        self.logger: QLogger = QueuedLogging.get_qlogger("logger_pool")
        self.logger.log(LogLevel.INFO, f"Initialized to run jobs on: {self.devices}")
        self.job_collection = JobCollection()
        if logger_collection_constructable is not None:
            logger_collection = logger_collection_constructable.construct()
            job_status_logger = JobStatusLogger(logger=logger_collection)
            subscriber = JobStatusLoggingSubscriber(job_status_logger)
            self.job_collection.add_subscriber(subscriber)

    def add_job(self, job: Job):
        self.job_q.put(job)
        self.job_collection.add_or_update_job(job)

    def add_jobs(self, jobs: List[Job]):
        self.logger.log(LogLevel.INFO, "Filling up job queue ... ")
        for job in tqdm.tqdm(jobs):
            self.job_q.put(job)
            self.job_collection.add_or_update_job(job)

    def run(self):
        # we have to add the termination jobs at the end of the queue such that the processes stop working and don't get stuck in jobs_q.get()
        termination_jobs = [Job(job_id=i+len(self.job_collection), fun=None, blueprint=None, param_dict=None,
                                job_type=JobType.TERMINATE) for i in range(self.num_processes)]
        self.add_jobs(termination_jobs)
        # create and start worker processes
        self.logger.log(LogLevel.INFO, f"Creating {self.num_processes} worker processes...")
        for process_id in tqdm.tqdm(range(self.num_processes)):
            self.create_or_replace_process(process_id, self.max_jobs_per_process)
        self.logger.log(LogLevel.INFO, f"Starting {self.num_processes} worker processes...")
        for p in tqdm.tqdm(self.worker_processes):
            p.start()
        # wait until all jobs are done
        while not self.job_collection.done:
            updated_job: Job = self.job_update_q.get()
            self.job_collection.add_or_update_job(updated_job)
            if updated_job.status == JobStatus.DONE and updated_job.error is not None:
                self.logger.log(
                    LogLevel.FATAL, f"FATAL! Job {updated_job.job_id} crashed while being executed by process {updated_job.executing_process_id} on {updated_job.device} after {int(updated_job.finishing_time - updated_job.starting_time)} seconds with error {updated_job.error}")
                self.logger.log(LogLevel.FATAL, f"Error: {updated_job.stacktrace}")
            elif updated_job.status == JobStatus.DONE:
                self.logger.log(
                    LogLevel.INFO, f"Job {updated_job.job_id} was successfully executed by process {updated_job.executing_process_id} on {updated_job.device} within {int(updated_job.finishing_time - updated_job.starting_time)} seconds.")
                self.logger.log(LogLevel.INFO, f"Progress: {int(self.job_collection.done_count / self.job_collection.job_count * 100)}%")
            if updated_job.status == JobStatus.DONE and updated_job.job_type == JobType.CALC:
                self.worker_processes[updated_job.executing_process_id].recreate_process_if_done()

    def create_or_replace_process(self, process_id: int, num_jobs_to_perform: int):
        process = WorkerProcessWrapper(process_id=process_id,
                                       num_jobs_to_perform=num_jobs_to_perform,
                                       device=self.devices[process_id % len(self.devices)],
                                       job_q=self.job_q,
                                       job_update_q=self.job_update_q)
        if len(self.worker_processes) == process.process_id:
            self.logger.log(LogLevel.DEBUG, f"Adding process {process_id}.")
            self.worker_processes.append(process)
        elif len(self.worker_processes) > process.process_id:
            self.worker_processes[process_id] = process
            self.logger.log(LogLevel.DEBUG, f"Replaced process {process_id}.")
        else:
            raise Exception("There is something wrong with the process id!")




import socketio
from typing import Dict, List


class ClientFactory:

    @staticmethod
    def get_buffered_client(client_id: str, host: str, port: int, disconnect_buffer_size: int, rooms: List[str]):
        sio_client = socketio.Client()
        bc = BufferedClient(client_id=client_id,
                            host=host,
                            port=port,
                            disconnect_buffer_size=disconnect_buffer_size,
                            sio_client=sio_client,
                            rooms=rooms)
        bc.connect()
        return bc


class BufferedClient:

    def __init__(self, client_id: str, host: str, port: int, disconnect_buffer_size: int, sio_client: socketio.Client, rooms: List[str]):
        self._client_id = client_id
        self._host = host
        self._port = port
        self._disconnect_buffer_size = disconnect_buffer_size
        self._sio_client = sio_client
        self.rooms = rooms

    @staticmethod
    def _register_callback_funs(sio_client: socketio.Client):
        # on message event
        sio_client.on("server_log_message", BufferedClient.on_server_log_message)
        sio_client.on("mlgym_event", BufferedClient.on_mlgym_event_message)

    def connect(self):
        self._sio_client.connect(f"{self._host}:{self._port}", wait=True, wait_timeout=20)
        BufferedClient._register_callback_funs(self._sio_client)
        self.emit("join", {"client_id": self._client_id, "rooms": [*self.rooms, self._client_id]})

    def leave(self):
        self.emit("leave", None)

    def on_server_log_message(data: Dict):
        print(data)

    def on_mlgym_event_message(data: Dict):
        print(data)
        # with open("log.txt", 'a') as fp:
        #     fp.write(json.dumps(data) + "\n")

    def emit(self, message_key: str,  message: Dict):
        self._sio_client.emit(message_key, message)
        # print(f"Sent message {message_key}: {message}")




from abc import abstractmethod
from ml_gym.early_stopping.early_stopping_strategies import EarlyStoppingIF
from ml_gym.models.nn.net import NNModel
from ml_gym.gym.evaluator import Evaluator
from ml_gym.gym.trainer import Trainer
from ml_gym.modes import RunMode
from ml_gym.optimizers.optimizer import OptimizerAdapter
import torch
from ml_gym.gym.stateful_components import StatefulComponent
from typing import List, Dict, Any
from ml_gym.util.logger import ConsoleLogger, LogLevel
from ml_gym.batching.batch import EvaluationBatchResult
from ml_gym.persistency.logging import ExperimentStatusLogger
from functools import partial
from ml_gym.persistency.io import GridSearchAPIClientIF, CheckpointResource
import pickle
from ml_gym.checkpointing.checkpointing import CheckpointingIF, CheckpointingInstruction


class AbstractGymJob(StatefulComponent):

    def __init__(self, experiment_status_logger: ExperimentStatusLogger):
        super().__init__()
        self._experiment_status_logger = experiment_status_logger
        # TODO: This is only a console logger that does not store on disk.
        # Since GymJob is a StatefulComponent we would have to implement serialization
        # for the queue in the QLogger. Solving this issue is left for the future
        self.logger = ConsoleLogger("logger_gym_job")

    @abstractmethod
    def execute(self, device: torch.device):
        raise NotImplementedError

    @staticmethod
    def from_blue_print(blue_print, device: torch.device) -> 'AbstractGymJob':
        return blue_print.construct(device)

    # def restore_state_in_stateful_components(self, measurement_id: int):
    #     state = DashifyReader.load_state(experiment_info=self.experiment_info, measurement_id=measurement_id)
    #     self.set_state(state)

    @staticmethod
    def batch_processed_callback(status: str, experiment_status_logger: ExperimentStatusLogger, num_batches: int,
                                 current_batch: int, splits: List[str], current_split: str, num_epochs: int, current_epoch: int):
        experiment_status_logger.log_experiment_status(status=status,
                                                       num_epochs=num_epochs,
                                                       current_epoch=current_epoch,
                                                       splits=splits,
                                                       current_split=current_split,
                                                       num_batches=num_batches,
                                                       current_batch=current_batch)

    @staticmethod
    def epoch_result_callback(experiment_status_logger: ExperimentStatusLogger, evaluation_result: EvaluationBatchResult,
                              current_epoch: int):
        experiment_status_logger.log_evaluation_results(evaluation_result, current_epoch)


class GymJob(AbstractGymJob):

    def __init__(self, grid_search_id: str, experiment_id: int,  run_mode: RunMode, model: NNModel, optimizer: OptimizerAdapter,
                 trainer: Trainer, evaluator: Evaluator, num_epochs: int, checkpointing_strategy: CheckpointingIF,
                 gs_api_client: GridSearchAPIClientIF, experiment_status_logger: ExperimentStatusLogger = None,
                 early_stopping_strategy: EarlyStoppingIF = None, warm_start_epoch: int = 0):
        super().__init__(experiment_status_logger)
        self.grid_search_id = grid_search_id
        self.experiment_id = experiment_id
        self.run_mode = run_mode
        self.model = model
        self.optimizer = optimizer
        # self.optimizer.register_model_params(dict(self.model.named_parameters()))
        self.num_epochs = num_epochs
        self.current_epoch = warm_start_epoch
        self.evaluator = evaluator
        self.trainer = trainer
        self.checkpointing_strategy = checkpointing_strategy
        self.early_stopping_strategy = early_stopping_strategy
        self.gs_api_client = gs_api_client
        if run_mode == RunMode.TRAIN:
            self._execution_method = self._execute_train
        elif run_mode == RunMode.WARM_START:
            self._execution_method = self._execute_warm_start
        else:
            raise NotImplementedError

    def _train_step(self, device: torch.device) -> NNModel:
        # self.restore_state_in_stateful_components(epoch - 1)
        # load model and optimizer
        # model_state = DashifyReader.load_model_state(self._experiment_info, epoch - 1)
        # self.model.load_state_dict(model_state)
        # self.model.to(device)
        # optimizer_state = DashifyReader.load_optimizer_state(self._experiment_info, epoch - 1)
        # self.optimizer.load_state_dict(optimizer_state)
        # train
        partial_batch_processed_callback = partial(self.batch_processed_callback, num_epochs=self.num_epochs,
                                                   current_epoch=self.current_epoch,
                                                   experiment_status_logger=self._experiment_status_logger)
        model = self.trainer.train_epoch(self.model, self.optimizer, device,
                                         batch_processed_callback_fun=partial_batch_processed_callback)
        return model

    def _evaluation_step(self, device: torch.device) -> List[EvaluationBatchResult]:
        self.model.to(device)
        partial_batch_processed_callback = partial(self.batch_processed_callback, num_epochs=self.num_epochs,
                                                   current_epoch=self.current_epoch,
                                                   experiment_status_logger=self._experiment_status_logger)
        partial_epoch_result_callback = partial(self.epoch_result_callback, current_epoch=self.current_epoch,
                                                experiment_status_logger=self._experiment_status_logger)

        evaluation_results = self.evaluator.evaluate(model=self.model,
                                                     device=device,
                                                     current_epoch=self.current_epoch,
                                                     num_epochs=self.num_epochs,
                                                     batch_processed_callback_fun=partial_batch_processed_callback,
                                                     epoch_result_callback_fun=partial_epoch_result_callback)

        return evaluation_results

    def run_checkpointing(self, checkpoint_instruction: CheckpointingInstruction):
        if checkpoint_instruction.save_current:
            self._experiment_status_logger.log_checkpoint(self.current_epoch, self.model.state_dict(), self.optimizer.state_dict(),
                                                          self.get_state())
        for epoch in checkpoint_instruction.checkpoints_to_delete:
            self._experiment_status_logger.log_checkpoint(epoch=epoch, model_binary_stream=None,
                                                          optimizer_binary_stream=None, stateful_components_binary_stream=None)

    def execute(self, device: torch.device):
        """ Executes the job

        Args:
            device: torch device either CPUs or a specified GPU
        """
        self._execution_method(device)

    def _execute_train(self, device: torch.device):
        self.optimizer.register_model_params(model_params=dict(self.model.named_parameters()))

        self.trainer.set_num_epochs(num_epochs=self.num_epochs)

        # initial evaluation
        evaluation_results = self._evaluation_step(device)

        # we store the initial model / last warmup model again
        checkpointing_instruction = self.checkpointing_strategy.get_model_checkpoint_instruction(num_epochs=self.num_epochs,
                                                                                                 current_epoch=self.current_epoch,
                                                                                                 evaluation_result=evaluation_results)
        self.run_checkpointing(checkpointing_instruction)

        # if early stopping criterion is fulfilled we can stop the training progress
        if self.early_stopping_strategy.is_stopping_criterion_fulfilled(current_epoch=self.current_epoch,
                                                                        evaluation_results=evaluation_results):
            return

        # start the actual training loop
        self.current_epoch += 1
        self.trainer.set_current_epoch(self.current_epoch)
        while not self.trainer.is_done():
            self.logger.log(LogLevel.INFO,  f"epoch: {self.current_epoch}")
            self._train_step(device)
            evaluation_results = self._evaluation_step(device)
            checkpointing_instruction = self.checkpointing_strategy.get_model_checkpoint_instruction(num_epochs=self.num_epochs,
                                                                                                     current_epoch=self.current_epoch,
                                                                                                     evaluation_result=evaluation_results)
            self.run_checkpointing(checkpointing_instruction)
            if self.early_stopping_strategy.is_stopping_criterion_fulfilled(current_epoch=self.current_epoch,
                                                                            evaluation_results=evaluation_results):
                break

            self.current_epoch += 1

    def _execute_warm_start(self, device: torch.device):
        if self.current_epoch > 0:
            model_state = pickle.loads(self.gs_api_client.get_checkpoint_resource(grid_search_id=self.grid_search_id,
                                                                                  experiment_id=self.experiment_id,
                                                                                  checkpoint_id=self.current_epoch,
                                                                                  checkpoint_resource=CheckpointResource.model))
            self.model.load_state_dict(model_state)

            optimizer_state = pickle.loads(self.gs_api_client.get_checkpoint_resource(grid_search_id=self.grid_search_id,
                                                                                      experiment_id=self.experiment_id,
                                                                                      checkpoint_id=self.current_epoch,
                                                                                      checkpoint_resource=CheckpointResource.optimizer))
            self.optimizer.load_state_dict(optimizer_state)

            state_component_state = pickle.loads(self.gs_api_client.get_checkpoint_resource(grid_search_id=self.grid_search_id,
                                                                                            experiment_id=self.experiment_id,
                                                                                            checkpoint_id=self.current_epoch,
                                                                                            checkpoint_resource=CheckpointResource.stateful_components))
            self.set_state(state_component_state)

        self._execute_train(device)

    def _execute_eval(self, device: torch.device):
        for epoch in self.num_epochs:
            # TODO need to load model + stateful components for the respective epoch here
            self._evaluation_step(device, epoch=epoch)


class GymJobFactory:
    @staticmethod
    def get_gym_job(grid_search_id: str, experiment_id: int, run_mode: RunMode, num_epochs: int, gs_api_client: GridSearchAPIClientIF,
                    experiment_status_logger: ExperimentStatusLogger = None, warm_start_epoch: int = 0,
                    **components: Dict[str, Any]) -> AbstractGymJob:
        return GymJob(grid_search_id=grid_search_id, experiment_id=experiment_id, run_mode=run_mode, num_epochs=num_epochs,
                      gs_api_client=gs_api_client, experiment_status_logger=experiment_status_logger,
                      warm_start_epoch=warm_start_epoch, **components)



from typing import Any, Type
from collections.abc import Mapping
from ml_gym.error_handling.exception import ClassRegistryKeyNotFoundError


class ClassRegistry(Mapping):
    """A decorated dictionary to act as a registry for models, loss functions etc."""

    def __init__(self):
        '''Use the object dict'''
        self._store = {}

    def __setitem__(self, key, value):
        self._store[key] = value

    def __getitem__(self, key) -> Type:
        return self._store[key]

    def __delitem__(self, key):
        del self._store[key]

    def __iter__(self):
        return iter(self._store)

    def __len__(self):
        return len(self._store)

    def __str__(self):
        '''returns simple dict representation of the mapping'''
        return str(self._store)

    def __repr__(self):
        '''echoes class, id, & reproducible representation in the REPL'''
        return '{}, ClassRegistry({})'.format(super(ClassRegistry, self).__repr__(), self._store)

    def add_class(self, key: str, element: Type):
        self._store[key] = element

    def get_instance(self, key: str, **params) -> Any:
        if key not in self._store:
            raise ClassRegistryKeyNotFoundError(f"Key {key} not present in class registry!")
        try:
            obj = self._store[key](**params)
        except Exception as e:
            raise Exception(f"Error building {self._store[key]}") from e
        return obj




from typing import List
from ml_gym.gym.stateful_components import StatefulComponent
from ml_gym.batching.batch import EvaluationBatchResult
from abc import abstractmethod
from dataclasses import dataclass, field


@dataclass
class CheckpointingInstruction:
    save_current: bool = False
    checkpoints_to_delete: List[int] = field(default_factory=list)


class CheckpointingIF(StatefulComponent):
    @abstractmethod
    def get_model_checkpoint_instruction(self, current_epoch: int, num_epochs: int,
                                         evaluation_result: EvaluationBatchResult) -> CheckpointingInstruction:
        raise NotImplementedError


class Checkpointing(CheckpointingIF):
    def __init__(self, checkpointing_strategy: CheckpointingIF):
        self.checkpointing_strategy = checkpointing_strategy

    def get_model_checkpoint_instruction(self, current_epoch: int, num_epochs: int,
                                         evaluation_result: EvaluationBatchResult) -> CheckpointingInstruction:
        return self.checkpointing_strategy.get_model_checkpoint_instruction(current_epoch=current_epoch, num_epochs=num_epochs,
                                                                            evaluation_result=evaluation_result)


class SaveLastEpochOnlyCheckpointingStrategy(CheckpointingIF):
    def __init__(self):
        pass

    def get_model_checkpoint_instruction(self, current_epoch: int, num_epochs: int,
                                         evaluation_result: EvaluationBatchResult) -> CheckpointingInstruction:
        return CheckpointingInstruction(save_current=True, checkpoints_to_delete=[current_epoch-1])


class SaveAllCheckpointingStrategy(CheckpointingIF):
    def __init__(self):
        pass

    def get_model_checkpoint_instruction(self, current_epoch: int, num_epochs: int,
                                         evaluation_result: EvaluationBatchResult) -> CheckpointingInstruction:
        return CheckpointingInstruction(save_current=True, checkpoints_to_delete=[])

# class RegularIntervalCheckpoint(Checkpointing):
#     def __init__(self,  tag: str, identifier: str,
#                  save_weights_only: Optional[bool] = False,
#                  checkpoint_interval: Optional[int] = None,
#                  save_last: bool = True):
#         super().__init__(tag=tag, identifier=identifier, save_weights_only=save_weights_only)
#         self.checkpoint_interval = checkpoint_interval
#         self.save_last = save_last

#     def model_checkpoint(self, current_epoch: int, num_epochs: int, evaluation_result: EvaluationBatchResult) -> Dict:
#         checkpoint_result = {'checkpoint_save': None,
#                              'checkpoint_delete': None}
#         if current_epoch == num_epochs:
#             if self.save_last:
#                 checkpoint_result['checkpoint_save'] = current_epoch
#         elif current_epoch % self.checkpoint_interval == 0:
#             checkpoint_result['checkpoint_save'] = current_epoch
#         self.checkpoint_history['epoch_' + str(current_epoch)] = checkpoint_result
#         return checkpoint_result


# class LastCheckpoint(Checkpointing):
#     def __init__(self,  tag: str, identifier: str,
#                  save_weights_only: Optional[bool] = False,
#                  save_last_only: bool = True
#                  ):
#         super().__init__(tag=tag, identifier=identifier, save_weights_only=save_weights_only)
#         self.save_last_only = save_last_only

#     def model_checkpoint(self,
#                          current_epoch: int,
#                          num_epochs: int) -> bool:
#         checkpoint_result = {'checkpoint_save': None,
#                              'checkpoint_delete': None}
#         if current_epoch == num_epochs:
#             checkpoint_result['checkpoint_save'] = current_epoch
#         self.checkpoint_history['epoch_' + str(current_epoch)] = checkpoint_result
#         return checkpoint_result


# class BestMetricCheckpoint(Checkpointing):
#     def __init__(self,  tag: str, identifier: str,
#                  inference_metric: str,
#                  mode: Optional[str] = None,
#                  best_count: int = 1,
#                  save_weights_only: Optional[bool] = False,

#                  ):
#         super().__init__(tag=tag, identifier=identifier, save_weights_only=save_weights_only)
#         self.mode = mode
#         self.inference_metric = inference_metric
#         self.best_count = best_count
#         self.best_metrics = {}

#     def model_checkpoint(self, current_epoch: int, num_epochs: int, evaluation_result: EvaluationBatchResult) -> Dict:

#         if not self.mode:
#             self.initialize_mode(evaluation_result=evaluation_result)
#         checkpoint_result = self.compare_best_metrics(current_epoch, evaluation_result)
#         self.checkpoint_history['epoch_' + str(current_epoch)] = checkpoint_result
#         return checkpoint_result

#     def compare_best_metrics(self, current_epoch: int, evaluation_result: EvaluationBatchResult) -> Dict:
#         checkpoint_result = {'checkpoint_save': None,
#                              'checkpoint_delete': None}
#         current_metric = self.get_current_metric(current_epoch, evaluation_result)
#         if len(self.best_metrics) < self.best_count:
#             self.best_metrics[current_epoch] = current_metric
#             checkpoint_result['checkpoint_save'] = current_epoch
#         else:
#             compare_fn = max if self.mode == "min" else min
#             best_item_key = compare_fn(self.best_metrics)
#             if compare_fn(self.best_metrics[best_item_key], current_metric) == current_metric:
#                 self.best_metrics.pop(best_item_key)
#                 self.best_metrics[current_epoch] = current_metric
#                 checkpoint_result['checkpoint_save'] = current_epoch
#                 checkpoint_result['checkpoint_delete'] = best_item_key

#         return checkpoint_result

#     def get_current_metric(self, current_epoch: int, evaluation_result: EvaluationBatchResult) -> float:
#         if self.inference_metric in evaluation_result.losses.keys():
#             current_metric = evaluation_result.losses[self.inference_metric][current_epoch]
#         elif self.inference_metric in evaluation_result.metrics.keys():
#             current_metric = evaluation_result.metrics[self.inference_metric][current_epoch]
#         else:
#             raise RuntimeError("{} not present in the metrics or the loss keys.".format(self.inference_metric))
#         return current_metric

#     def initialize_mode(self, evaluation_result: EvaluationBatchResult):
#         if not self.mode:
#             # set up min-max mode based on the metric type if not specified by the user
#             if self.inference_metric in evaluation_result.losses.keys():
#                 self.mode = "min"
#             elif self.inference_metric in evaluation_result.metrics.keys():
#                 # TODO: check if metrics are always maximized or use some better logic to set mode
#                 self.mode = "max"
#             else:
#                 raise RuntimeError("{} not present in the metrics or the loss keys.".format(self.inference_metric))



from abc import ABC, abstractmethod
from typing import Dict, Any, Type, List
from ml_gym.blueprints.blue_prints import BluePrint
from ml_gym.persistency.io import GridSearchAPIClientConstructableIF


class ValidatorIF(ABC):

    @abstractmethod
    def create_blueprints(self, grid_search_id: str, blue_print_type: Type[BluePrint], gs_config: Dict[str, Any],
                          num_epochs: int, dashify_logging_path: str,
                          gs_api_client_constructable: GridSearchAPIClientConstructableIF) -> List[BluePrint]:
        raise NotImplementedError



from fastapi import FastAPI
from fastapi import status, HTTPException
from fastapi.responses import StreamingResponse
from ml_board.backend.restful_api.data_access import DataAccessIF
from ml_gym.error_handling.exception import InvalidPathError
from ml_board.backend.restful_api.data_models import RawTextFile, CheckpointResource
from typing import Callable
# from fastapi.staticfiles import StaticFiles


class RestfulAPIServer:
    def __init__(self, data_access: DataAccessIF):
        self.app = FastAPI(port=8080)
        self.data_access = data_access
        self.app.add_api_route(path="/grid_searches/{grid_search_id}/experiments",
                               methods=["GET"], endpoint=self.get_experiment_statuses)
        self.app.add_api_route(path="/grid_searches/{grid_search_id}/{config_name}",
                               methods=["PUT"], endpoint=self.add_raw_config_to_grid_search)
        self.app.add_api_route(path="/grid_searches/{grid_search_id}/{experiment_id}/{config_name}",
                               methods=["PUT"], endpoint=self.add_config_to_experiment)
        self.app.add_api_route(path="/checkpoints/{grid_search_id}/{experiment_id}/{epoch}/{checkpoint_resource}",
                               methods=["GET"], endpoint=self.get_checkpoint_resource)
        self.app.add_api_route(path="/checkpoints/{grid_search_id}/{experiment_id}/{epoch}",
                               methods=["GET"], endpoint=self.get_checkpoint_dict_epoch)

        # self.app.mount("/", StaticFiles(directory="/home/mluebberin/repositories/github/private_workspace/mlgym/src/ml_board/frontend/dashboard/build/", html=True), name="static")

    def get_experiment_statuses(self, grid_search_id: str):
        try:
            experiment_statuses = self.data_access.get_experiment_statuses(grid_search_id)
            return experiment_statuses
        except InvalidPathError as e:
            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN,
                                detail=f'Provided invalid grid_search_id {grid_search_id}') from e

    # @app.get('/grid_searches/{grid_search_id}/gs_config')
    # def get_grid_search_config(grid_search_id: str):
    #     requested_full_path = os.path.realpath(os.path.join(top_level_logging_path, str(grid_search_id), "gs_config.yml"))

    #     if is_safe_path(base_dir=top_level_logging_path, requested_path=requested_full_path):
    #         return YAMLConfigLoader.load(requested_full_path)

    #     else:
    #         raise HTTPException(status_code=status.HTTP_403_FORBIDDEN,
    #                             detail=f'Provided unsafe filepath {requested_full_path}')

    def add_raw_config_to_grid_search(self, grid_search_id: str, config_name: str, config_file: RawTextFile):
        try:
            self.data_access.add_raw_config_to_grid_search(grid_search_id=grid_search_id,
                                                           config_name=config_name,
                                                           config_file=config_file)
        except InvalidPathError as e:
            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN,
                                detail=f'Provided invalid grid_search_id {grid_search_id} or config_name {config_name}') from e

    def get_raw_config_of_grid_search(self, grid_search_id: str, config_name: str):
        raise NotImplementedError

    def add_config_to_experiment(self, grid_search_id: str, experiment_id: str, config_name: str, config: RawTextFile):
        try:
            self.data_access.add_config_to_experiment(grid_search_id=grid_search_id,
                                                      experiment_id=experiment_id,
                                                      config_name=config_name,
                                                      config=config)
        except InvalidPathError as e:
            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN,
                                detail=f'Provided invalid grid_search_id {grid_search_id}, experiment_id {experiment_id} or config_name {config_name}') from e

    def get_checkpoint_resource(self, grid_search_id: str, experiment_id: str, epoch: str, checkpoint_resource: CheckpointResource):
        try:
            file_generator = self.data_access.get_checkpoint_resource(grid_search_id=grid_search_id,
                                                                      experiment_id=experiment_id,
                                                                      epoch=epoch,
                                                                      checkpoint_resource=checkpoint_resource)
            response = StreamingResponse(file_generator, media_type="application/octet-stream")
            return response
        except InvalidPathError as e:
            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN,
                                detail="Provided invalid parameters for checkpoint resource.") from e

    def get_checkpoint_dict_epoch(self, grid_search_id: str, experiment_id: str, epoch: str):
        try:
            checkpoint_list = self.data_access.get_checkpoint_dict_epoch(grid_search_id=grid_search_id,
                                                                         experiment_id=experiment_id,
                                                                         epoch=epoch)
            return checkpoint_list
        except InvalidPathError as e:
            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN,
                                detail=f'Provided invalid grid_search_id {grid_search_id}, experiment_id {experiment_id} or epoch {epoch}') from e

    def run_server(self, application_server_callable: Callable):
        application_server_callable(app=self.app)



from ml_gym.persistency.logging import JobStatusLogger, MLgymStatusLoggerCollectionConstructable
import torch
from typing import Dict, List
from ml_gym.multiprocessing.pool import Pool, Job
from ml_gym.blueprints.blue_prints import BluePrint
from ml_gym.gym.jobs import AbstractGymJob
from ml_gym.util.devices import get_devices
import tqdm


torch.multiprocessing.set_start_method('spawn', force=True)


class Gym:
    def __init__(self, job_id_prefix: str, process_count: int = 1, device_ids: List[int] = None, log_std_to_file: bool = True,
                 logger_collection_constructable: MLgymStatusLoggerCollectionConstructable = None):
        self.devices = get_devices(device_ids)
        self.logger_collection_constructable = logger_collection_constructable
        self.job_status_logger = JobStatusLogger(self.logger_collection_constructable.construct())
        self.log_std_to_file = log_std_to_file
        self.pool = Pool(num_processes=process_count, devices=self.devices, logger_collection_constructable=logger_collection_constructable)
        self.jobs: List[Job] = []
        self.job_id_prefix = job_id_prefix
        self.job_counter = 0

    def run(self, parallel=True):
        """Executes the jobs. Note that this function blocks until all jobs have been executed!

        Args:
            parallel (bool, optional): When set to True, jobs are run in parallel in the multiprocessing environment. Defaults to True.
        """
        if parallel:
            for _ in range(len(self.jobs)):
                job = self.jobs.pop(0)
                self.pool.add_job(job)

            self.pool.run()
        else:
            for _ in tqdm.tqdm(range(len(self.jobs)), desc="Models trained"):
                job = self.jobs.pop(0)
                self.work(job, self.devices[0])

    def add_blueprint(self, blueprint: BluePrint) -> int:
        job = Job(job_id=f"{blueprint.grid_search_id}-{self.job_counter}", fun=Gym._run_job, blueprint=blueprint, param_dict={"log_std_to_file": self.log_std_to_file})
        self.job_counter += 1
        self.jobs.append(job)
        return job.job_id

    def add_blueprints(self, blueprints: List[BluePrint]):
        for blueprint in blueprints:
            job_id = self.add_blueprint(blueprint)
            self.job_status_logger.log_experiment_config(grid_search_id=blueprint.grid_search_id,
                                                         experiment_id=blueprint.experiment_id,
                                                         job_id=job_id,
                                                         config=blueprint.config)

    @staticmethod
    def _run_job(blueprint: BluePrint, device: torch.device, log_std_to_file: bool) -> AbstractGymJob:
        gym_job = AbstractGymJob.from_blue_print(blueprint, device=device)
        return gym_job.execute(device=device)

    def work(self, job: Job, device: torch.device):
        job.device = device
        job.execute()



from multiprocessing import Queue, Process
from ml_board.backend.messaging.events import Event
from abc import ABC
from typing import List, Dict
from ml_board.backend.messaging.event_storage import EventStorageIF


class PublishingIF(ABC):

    def publish_event(self, event: Event):
        raise NotImplementedError


class BrokerIF(ABC):
    def listen_for_events(self, event_queue: Queue):
        raise NotImplementedError


class SubscriberIF:

    def update_event(self, event: Event):
        raise NotImplementedError


class Broker(Process, BrokerIF):

    def __init__(self, event_storage: EventStorageIF, event_queue: Queue):
        super().__init__(target=self.work, args=(event_queue, ))

        self._event_storage: EventStorageIF = event_storage
        self._event_queue = event_queue
        # maps subscriber to the last event id
        self._subscribers: Dict[SubscriberIF, int] = {}

    def listen_for_events(self, event_queue: Queue):
        while True:
            event = event_queue.get()
            self._publish_event(event)

    def _update_subscribers(self):
        for subscriber, last_event_id in self._subscribers.items():
            while self._event_storage.has_next(last_event_id):
                last_event_id, event = self._event_storage.get_next(last_event_id)
                subscriber.update_event(event)
                self._subscribers[subscriber] = last_event_id

    def _publish_event(self, event: Event):
        self._message_queue.put(event)
        self._update_subscribers()

    def add_subscriber(self, subscriber: SubscriberIF):
        self._subscribers[subscriber] = 0


class Publisher(PublishingIF):
    def __init__(self, event_queue: Queue):
        self._event_queue = event_queue

    def publish_event(self, event: Event):
        self._event_queue.put(event)



from flask import Flask, render_template, request, copy_current_request_context
from flask_socketio import SocketIO, emit, join_room, rooms, disconnect
from typing import List
from collections import defaultdict


class EventSubscriberIF:

    def callback(self):
        raise NotImplementedError


class WebSocketServer:

    def __init__(self, host: str, port: int, message_delay: int, log_file_path: str, async_mode: str, app: Flask):
        self._host = host
        self._port = port
        self._message_delay = message_delay
        self._socketio = SocketIO(app, async_mode=async_mode, cors_allowed_origins=["http://localhost:3000", "http://localhost:7000"])
        self._client_sids = []
        self._init_call_backs()

        # load the log file
        with open(log_file_path, encoding="utf-8") as fp:
            self.log_list = fp.readlines()

    def run(self, app: Flask):
        self._socketio.run(app, host=self._host, port=self._port)

    def emit_server_log_message(self, data):
        emit("server_log_message", data)

    @property
    def client_sids(self) -> List[str]:
        return self._client_sids

    def _init_call_backs(self):

        @self._socketio.on("join")
        def on_join(data):

            @copy_current_request_context
            def _send_event_history_to_client(log_list):
                for msg in log_list:
                    emit('mlgym_event', msg)
                    self._socketio.sleep(self._message_delay)

            # save the sids of the new client
            client_sid = request.sid
            self._client_sids.append(client_sid)
            if 'client_id' in data:
                client_id = data['client_id']
            else:
                client_id = "<unknown>"
            # join the rooms we want to subcribe to
            rooms_to_join = data['rooms']
            for room in rooms_to_join:
                join_room(room)
            print(f"Client {client_id} joined rooms: {rooms()}")
            self.emit_server_log_message(f"Client {client_id} joined rooms: {rooms()}")

            # send the history of log messages to the client
            if "mlgym_event_subscribers" in rooms_to_join:
                self._socketio.start_background_task(lambda: _send_event_history_to_client(self.log_list))

        @self._socketio.on("leave")
        def on_leave():
            self._client_sids.remove(request.sid)
            # TODO  leave all rooms
            # leave_room(message['room'])
            self.emit_server_log_message("You are now disconnected.")
            disconnect()

        @self._socketio.on("ping")
        def on_ping():
            emit('pong')

        @self._socketio.on("client_connected")
        def on_client_connected():
            print(f"Client with SID {request.sid} connnected.")
            self.emit_server_log_message(f"Client with SID {request.sid} connnected.")

        @self._socketio.on("client_disconnected")
        def on_client_disconnected():
            print('Client disconnected', request.sid)
            self._client_sids.remove(request.sid)


if __name__ == '__main__':
    async_mode = None

    app = Flask(__name__, template_folder="template")
    app.config['SECRET_KEY'] = 'secret!'

    port = 7000
    host = "localhost"
    async_mode = None

    log_file_path = "/home/mluebberin/repositories/github/private_workspace/mlgym/src/ml_gym/backend/streaming/log.txt"

    message_delay = 0.1  # in seconds

    ws = WebSocketServer(host=host, port=port, message_delay=message_delay, log_file_path=log_file_path, async_mode=async_mode, app=app)

    @app.route('/')
    def index():
        return render_template('index.html', async_mode=ws._socketio.async_mode)

    # @app.route('/status')
    # def status():
    #     return render_template('status.html', async_mode=ws._socketio.async_mode)

    @app.route('/api/status')
    def api_status():
        client_sids = ws.client_sids
        room_key_to_sid = defaultdict(list)
        for client_sid in client_sids:
            room_keys = rooms(client_sid, "/")
            for room_key in room_keys:
                room_key_to_sid[room_key].append(client_sid)

        return {"clients": client_sids, "rooms": room_key_to_sid}

    ws.run(app)



from pydantic import BaseModel
from enum import Enum
from typing import Dict


class FileFormat(str, Enum):
    YAML = "YAML"
    JSON = "JSON"


class RawTextFile(BaseModel):
    file_format: FileFormat
    content: str


class ExperimentStatus(BaseModel):
    experiment_id: int
    last_checkpoint_id: int
    experiment_config: Dict


class CheckpointResource(str, Enum):
    model = "model"
    optimizer = "optimizer"
    stateful_components = "stateful_components"



from typing import Dict, Any, Type, List
from ml_gym.blueprints.blue_prints import BluePrint
from ml_gym.modes import RunMode
from ml_gym.persistency.logging import MLgymStatusLoggerCollectionConstructable
from ml_gym.validation.validator import ValidatorIF
from ml_gym.util.grid_search import GridSearch
from ml_gym.persistency.io import GridSearchAPIClientConstructableIF


class GridSearchValidator(ValidatorIF):
    def __init__(self, run_mode: RunMode):
        self.run_mode = run_mode

    def create_blueprints(self, grid_search_id: str, blue_print_type: Type[BluePrint], gs_config: Dict[str, Any], num_epochs: int,
                          gs_api_client_constructable: GridSearchAPIClientConstructableIF,
                          logger_collection_constructable: MLgymStatusLoggerCollectionConstructable = None) -> List[BluePrint]:
        run_id_to_config_dict = {run_id: config for run_id, config in enumerate(GridSearch.create_gs_from_config_dict(gs_config))}

        blueprints = []
        for config_id, experiment_config in run_id_to_config_dict.items():
            blueprint = BluePrint.create_blueprint(blue_print_class=blue_print_type,
                                                   run_mode=self.run_mode,
                                                   experiment_config=experiment_config,
                                                   num_epochs=num_epochs,
                                                   grid_search_id=grid_search_id,
                                                   experiment_id=config_id,
                                                   logger_collection_constructable=logger_collection_constructable,
                                                   gs_api_client_constructable=gs_api_client_constructable)
            blueprints.append(blueprint)
        return blueprints



from abc import abstractmethod
from ml_gym.blueprints.blue_prints import BluePrint
from ml_gym.multiprocessing.states import JobStatus, JobType
import torch
from typing import Callable, Dict, List


class JobIF:
    @property
    def device(self) -> torch.device:
        raise NotImplementedError

    @property
    def experiment_id(self) -> str:
        raise NotImplementedError

    @property
    def grid_search_id(self) -> str:
        raise NotImplementedError

    @device.setter
    def device(self, value: torch.device):
        raise NotImplementedError

    def execute(self):
        raise NotImplementedError


class Job(JobIF):
    def __init__(self, job_id: str, fun: Callable, blueprint: BluePrint, param_dict: Dict, job_type: JobType = JobType.CALC):
        self.job_id = job_id
        self.job_type = job_type
        self.fun = fun
        self.blueprint = blueprint
        self.param_dict = param_dict
        self.status = JobStatus.INIT
        self.starting_time = -1
        self.finishing_time = -1
        self.executing_process_id = -1
        self.error = None
        self.stacktrace = None
        self._device: torch.device = ""

    @property
    def device(self) -> torch.device:
        return self._device

    @property
    def experiment_id(self) -> str:
        if self.blueprint is not None:
            return self.blueprint.experiment_id
        else:
            return None

    @property
    def grid_search_id(self) -> str:
        if self.blueprint is not None:
            return self.blueprint.grid_search_id
        else:
            return None

    @device.setter
    def device(self, value: torch.device):
        self._device = value

    def execute(self):
        self.param_dict["device"] = self._device
        return self.fun(blueprint=self.blueprint, **self.param_dict)


class JobStatusSubscriberIF:

    @abstractmethod
    def callback_job_event(self, job: Job):
        raise NotImplementedError


class JobCollection:

    def __init__(self):
        self.job_dict: Dict[str, Job] = {}
        self.subscribers: List[JobStatusSubscriberIF] = []

    def add_or_update_job(self, job: Job):
        self.job_dict[job.job_id] = job
        if job.job_type == JobType.CALC:
            self.update_subscribers(job)

    def add_subscriber(self, subscriber: JobStatusSubscriberIF):
        self.subscribers.append(subscriber)

    def update_subscribers(self, job: Job):
        for s in self.subscribers:
            s.callback_job_event(job)

    def __len__(self) -> int:
        return len([1 for job in self.job_dict.values() if job.job_type == JobType.CALC])

    @property
    def done(self) -> bool:
        return all([job.status == JobStatus.DONE for job in self.job_dict.values() if job.job_type == JobType.CALC])

    @property
    def done_count(self) -> bool:
        return sum([job.status == JobStatus.DONE for job in self.job_dict.values() if job.job_type == JobType.CALC])

    @property
    def job_count(self) -> int:
        return len([1 for job in self.job_dict.values() if job.job_type == JobType.CALC])



from typing import Dict, Any, Type
from ml_gym.modes import RunMode, ValidationMode
from ml_gym.validation.nested_cross_validation import NestedCV
from ml_gym.validation.cross_validation import CrossValidation
from ml_gym.validation.gs_validator import GridSearchValidator, ValidatorIF
from ml_gym.blueprints.blue_prints import BluePrint
from ml_gym.error_handling.exception import ValidationModeNotValidError


class ValidatorFactory:

    @staticmethod
    def get_nested_cv(gs_config: Dict[str, Any], cv_config: Dict[str, Any], blue_print_type: Type[BluePrint],
                      run_mode: RunMode) -> NestedCV:
        iterator_key = cv_config["NESTED_CROSS_VALIDATION"]["iterator_key"]
        split_key = cv_config["NESTED_CROSS_VALIDATION"]["split_key"]
        component_names = [iterator_key]
        components = blue_print_type.construct_components(config=gs_config, component_names=component_names)
        iterator = components[iterator_key][split_key]
        return NestedCV(dataset_iterator=iterator, **cv_config["NESTED_CROSS_VALIDATION"]["config"], run_mode=run_mode)

    @staticmethod
    def get_cross_validator(gs_config: Dict[str, Any], cv_config: Dict[str, Any], blue_print_type: Type[BluePrint],
                            run_mode: RunMode) -> CrossValidation:
        iterator_key = cv_config["CROSS_VALIDATION"]["iterator_key"]
        split_key = cv_config["CROSS_VALIDATION"]["split_key"]
        component_names = [iterator_key]
        components = blue_print_type.construct_components(config=gs_config, component_names=component_names)
        iterator = components[iterator_key][split_key]
        return CrossValidation(dataset_iterator=iterator, **cv_config["CROSS_VALIDATION"]["config"], run_mode=run_mode)

    @staticmethod
    def get_gs_validator(run_mode: RunMode) -> GridSearchValidator:
        return GridSearchValidator(run_mode=run_mode)


def get_validator(validation_mode: ValidationMode, blue_print_class:  Type[BluePrint],
                  run_mode: RunMode, cv_config: Dict = None, gs_config: Dict = None,) -> ValidatorIF:
    if validation_mode == ValidationMode.GRID_SEARCH:
        validator = ValidatorFactory.get_gs_validator(run_mode=run_mode)
    elif validation_mode == ValidationMode.CROSS_VALIDATION:
        validator = ValidatorFactory.get_cross_validator(gs_config=gs_config,
                                                         cv_config=cv_config,
                                                         blue_print_type=blue_print_class,
                                                         run_mode=run_mode)
    elif validation_mode == ValidationMode.NESTED_CV:
        validator = ValidatorFactory.get_nested_cv(gs_config=gs_config,
                                                   cv_config=cv_config,
                                                   blue_print_type=blue_print_class,
                                                   run_mode=run_mode)
    else:
        raise ValidationModeNotValidError
    return validator



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, List, Dict
from ml_gym.multiprocessing.states import JobStatus, JobType
from ml_gym.io.websocket_client import ClientFactory, BufferedClient
#from ml_gym.gym.model_checkpointing import ModelCheckpointComponent
import time
import torch
import pickle


def get_timestamp() -> int:
    return int(time.time_ns()/1000)  # in ms


class MLgymStatusLoggerIF(ABC):

    # @abstractmethod
    # def log_job_status(self, job_id: int, job_type: JobType, status: JobStatus, experiment_id: str, starting_time: int, finishing_time: int, device: torch.device,
    #                    error: str = "", stacktrace: str = ""):
    #     raise NotImplementedError

    # @abstractmethod
    # def log_model_status(self, experiment_id: str, status: str, current_epoch: int, splits: List[str], current_split: str, split_progress: int):
    #     raise NotImplementedError

    # @abstractmethod
    # def log_evaluation_results(self, eval_result: EvaluationBatchResult, epoch: int, experiment_id: str):
    #     raise NotImplementedError

    @abstractmethod
    def log_raw_message(self, raw_log_message: Dict):
        raise NotImplementedError


class LoggerCollection(MLgymStatusLoggerIF):

    def __init__(self, loggers: MLgymStatusLoggerIF):
        self._loggers = loggers

    # def log_job_status(self, job_id: int, job_type: JobType, status: JobStatus, experiment_id: str, starting_time: int, finishing_time: int, device: torch.device,
    #                    error: str = "", stacktrace: str = ""):
    #     for logger in self._loggers:
    #         logger.log_job_status(job_id, job_type, status, experiment_id, starting_time, finishing_time, device, error, stacktrace)

    # def log_model_status(self, experiment_id: str, status: str, current_epoch: int, splits: List[str], current_split: str, split_progress: int):
    #     for logger in self._loggers:
    #         logger.log_model_status(experiment_id, status, current_epoch, splits, current_split, split_progress)

    # def log_evaluation_results(self, eval_result: EvaluationBatchResult, epoch: int, experiment_id: str):
    #     for logger in self._loggers:
    #         logger.log_evaluation_results(eval_result, epoch, experiment_id)

    def log_raw_message(self, raw_log_message: Dict):
        for logger in self._loggers:
            logger.log_raw_message(raw_log_message)


class DiscLogger(MLgymStatusLoggerIF):
    def __init__(self):
        pass

    # # def log_evaluation_results(self, results: List[EvaluationBatchResult], epoch: int, experiment_info: ExperimentInfo):
    # #     DashifyWriter.log_measurement_result(results=results, experiment_info=experiment_info, measurement_id=epoch)

    # def log_raw_message(self, raw_log_message: Dict):
    #     # TODO FIX experiment_info in interface
    #     DashifyWriter.log_raw_experiment_message(data_dict=raw_log_message, experiment_info=None)


class StreamedLogger(MLgymStatusLoggerIF):
    def __init__(self, host: str, port: int):
        self._sio_client: BufferedClient = ClientFactory.get_buffered_client(client_id="mlgym_streamed_logger",
                                                                             host=host,
                                                                             port=port,
                                                                             disconnect_buffer_size=0,
                                                                             rooms=["mlgym_event_publishers"])

    def log_raw_message(self, raw_log_message: Dict):
        self._sio_client.emit(message_key="mlgym_event", message=raw_log_message)


class JobStatusLogger:
    def __init__(self, logger: MLgymStatusLoggerIF) -> None:
        self._logger = logger

    def log_job_status(self, job_id: str, job_type: JobType, status: JobStatus, grid_search_id: str, experiment_id: str, starting_time: int, finishing_time: int,
                       device: torch.device, error: str = "", stacktrace: str = ""):
        message = {"event_type": "job_status", "creation_ts": get_timestamp()}
        payload = {"job_id": job_id, "job_type": job_type.value, "status": status.value, "grid_search_id": grid_search_id, "experiment_id": experiment_id,
                   "starting_time": starting_time, "finishing_time": finishing_time, "device": str(device), "error": error,
                   "stacktrace": stacktrace}
        message["payload"] = payload
        self._logger.log_raw_message(raw_log_message=message)

    def log_experiment_config(self, grid_search_id: str, experiment_id: str, job_id: str, config: Dict[str, Any]):
        message = {"event_type": "experiment_config", "creation_ts": get_timestamp()}
        payload = {"grid_search_id": grid_search_id, "experiment_id": experiment_id, "job_id": job_id, "config": config}
        message["payload"] = payload
        self._logger.log_raw_message(raw_log_message=message)


class ExperimentStatusLogger:
    def __init__(self, logger: MLgymStatusLoggerIF, experiment_id: str, grid_search_id: str) -> None:
        self._logger = logger
        self._grid_search_id = grid_search_id
        self._experiment_id = experiment_id

    def log_experiment_status(self, status: str, num_epochs: int, current_epoch: int, splits: List[str], current_split: str,
                              num_batches: int, current_batch: int):
        message = {"event_type": "experiment_status", "creation_ts": get_timestamp()}
        payload = {"grid_search_id": self._grid_search_id, "experiment_id": self._experiment_id, "status": status, "num_epochs": num_epochs, "current_epoch": current_epoch,
                   "splits": splits, "current_split": current_split, "num_batches": num_batches, "current_batch": current_batch}
        message["payload"] = payload
        self._logger.log_raw_message(raw_log_message=message)

    def log_evaluation_results(self, eval_result: "EvaluationBatchResult", epoch: int):
        message = {"event_type": "evaluation_result", "creation_ts": get_timestamp()}
        metric_scores = [{"metric": metric_key, "split": eval_result.split_name, "score": metric_score[0]}
                         for metric_key, metric_score in eval_result.metrics.items()]
        loss_scores = [{"loss": loss_key, "split": eval_result.split_name, "score": loss_score[0]}
                       for loss_key, loss_score in eval_result.losses.items()]
        payload = {"grid_search_id": self._grid_search_id, "experiment_id": self._experiment_id, "epoch": epoch}
        payload["metric_scores"] = metric_scores
        payload["loss_scores"] = loss_scores
        message["payload"] = payload
        self._logger.log_raw_message(raw_log_message=message)

    def log_checkpoint(self, epoch: int, model_binary_stream=None, optimizer_binary_stream=None, stateful_components_binary_stream=None):
        message = {"event_type": "checkpoint", "creation_ts": get_timestamp()}
        payload = {
            "grid_search_id": self._grid_search_id,
            "experiment_id": self._experiment_id,
            "checkpoint_id": epoch,
            "checkpoint_streams": {
                "model": pickle.dumps(model_binary_stream) if model_binary_stream is not None else None,
                "optimizer": pickle.dumps(optimizer_binary_stream) if optimizer_binary_stream is not None else None,
                "stateful_components": pickle.dumps(stateful_components_binary_stream) if stateful_components_binary_stream is not None else None
            }
        }
        message["payload"] = payload
        # for i in range(1000):
        self._logger.log_raw_message(raw_log_message=message)


class MLgymStatusLoggerTypes(Enum):
    DISC_LOGGER = DiscLogger
    STREAMED_LOGGER = StreamedLogger


@dataclass
class MLgymStatusLoggerConfig:
    logger_type: MLgymStatusLoggerTypes
    logger_config: Dict


class LoggerConstructableIF:

    @abstractmethod
    def construct(self) -> MLgymStatusLoggerIF:
        raise NotImplementedError


@dataclass
class MLgymStatusLoggerConstructable(LoggerConstructableIF):
    config: MLgymStatusLoggerConfig

    def construct(self) -> MLgymStatusLoggerIF:
        return self.config.logger_type.value(**self.config.logger_config)


@dataclass
class MLgymStatusLoggerCollectionConstructable(LoggerConstructableIF):
    configs: List[MLgymStatusLoggerConfig]

    def construct(self) -> MLgymStatusLoggerIF:
        loggers = [MLgymStatusLoggerConstructable(config).construct() for config in self.configs]
        return LoggerCollection(loggers)



from abc import ABC, abstractmethod
import os
import glob
import re
from typing import Dict, List
from ml_gym.error_handling.exception import InvalidPathError
import json
from ml_board.backend.restful_api.data_models import RawTextFile, CheckpointResource, ExperimentStatus


class DataAccessIF(ABC):

    @abstractmethod
    def get_experiment_statuses(self, grid_search_id: str):
        raise NotImplementedError

    @abstractmethod
    def add_raw_config_to_grid_search(self, grid_search_id: str, config_name: str, config_file: RawTextFile):
        raise NotImplementedError

    @abstractmethod
    def get_raw_config_of_grid_search(self, grid_search_id: str, config_name: str):
        raise NotImplementedError

    @abstractmethod
    def add_config_to_experiment(self, grid_search_id: str, experiment_id: str, config_name: str, config: RawTextFile):
        raise NotImplementedError

    @abstractmethod
    def get_checkpoint_resource(self, grid_search_id: str, experiment_id: str, epoch: str, checkpoint_resource: CheckpointResource):
        raise NotImplementedError

    @abstractmethod
    def get_checkpoint_dict_epoch(self, grid_search_id: str, experiment_id: str, epoch: str):
        raise NotImplementedError


class FileDataAccess(DataAccessIF):

    def __init__(self, top_level_logging_path: str):
        print(top_level_logging_path)
        self.top_level_logging_path = top_level_logging_path

    @staticmethod
    def is_safe_path(base_dir, requested_path, follow_symlinks=True):
        # resolves symbolic links
        if follow_symlinks:
            matchpath = os.path.realpath(requested_path)
        else:
            matchpath = os.path.abspath(requested_path)
        return base_dir == os.path.commonpath((base_dir, matchpath))

    @staticmethod
    def get_checkpoint_files(path: str, base_path: str):
        full_paths = glob.glob(os.path.join(path, "**", "*.bin"), recursive=True)
        return [os.path.relpath(full_path, base_path) for full_path in full_paths]

    @staticmethod
    def iterfile(file_path: str):
        with open(file_path, mode="rb") as file_like:
            yield from file_like

    def get_experiment_statuses(self, grid_search_id: str) -> List[ExperimentStatus]:
        def get_last_checkpoint_ids(top_level_logging_path: str, grid_search_id: str) -> Dict:
            paths = glob.glob(os.path.join(top_level_logging_path, grid_search_id, '**/*'), recursive=True)
            regex_checkpoints = r".*-\d\d\/(\d)\/(\d)$"
            regex_experiments = r".*-\d\d\/(\d)$"

            experiment_id_to_checkpoint_ids = {}
            for path in paths:
                # experiment match
                match = re.match(regex_experiments, path)
                if match is not None:
                    experiment_id = int(match.groups()[0])
                    if experiment_id not in experiment_id_to_checkpoint_ids:
                        experiment_id_to_checkpoint_ids[experiment_id] = [-1]

                # checkpoint match
                match = re.match(regex_checkpoints, path)
                if match is not None:
                    experiment_id, checkpoint_id = int(match.groups()[0]), int(match.groups()[1])
                    if experiment_id not in experiment_id_to_checkpoint_ids:
                        experiment_id_to_checkpoint_ids[experiment_id] = []
                    experiment_id_to_checkpoint_ids[experiment_id].append(checkpoint_id)
            epxeriment_id_to_last_checkpoint = {k: max(v) for k, v in experiment_id_to_checkpoint_ids.items()}
            return epxeriment_id_to_last_checkpoint

        def load_experiment_config(top_level_logging_path: str, grid_search_id: str, experiment_id: int) -> Dict:
            full_path = os.path.join(top_level_logging_path, grid_search_id, str(experiment_id), "experiment_config.json")
            with open(full_path, "r") as fp:
                experiment_config = json.load(fp)
            return experiment_config

        requested_full_path = os.path.realpath(os.path.join(self.top_level_logging_path, grid_search_id))

        if FileDataAccess.is_safe_path(base_dir=self.top_level_logging_path, requested_path=requested_full_path):
            epxeriment_id_to_last_checkpoint_id = get_last_checkpoint_ids(self.top_level_logging_path, grid_search_id)
            response = [ExperimentStatus(**{"experiment_id": experiment_id,
                                            "last_checkpoint_id": last_checkpoint_id,
                                            "experiment_config": load_experiment_config(self.top_level_logging_path, grid_search_id,
                                                                                        experiment_id)})
                        for experiment_id, last_checkpoint_id in epxeriment_id_to_last_checkpoint_id.items()]
            return response

        else:
            raise InvalidPathError(f"File path {requested_full_path} is not safe.")

    def add_raw_config_to_grid_search(self, grid_search_id: str, config_name: str, config_file: RawTextFile):
        requested_full_path = os.path.realpath(os.path.join(self.top_level_logging_path, str(grid_search_id), config_name))

        if FileDataAccess.is_safe_path(base_dir=self.top_level_logging_path, requested_path=requested_full_path):
            os.makedirs(os.path.dirname(requested_full_path), exist_ok=True)
            with open(requested_full_path, "w") as fp:
                fp.writelines(config_file.content)
        else:
            raise InvalidPathError(f"File path {requested_full_path} is not safe.")

    def get_raw_config_of_grid_search(self, grid_search_id: str, config_name: str):
        raise NotImplementedError

    def add_config_to_experiment(self, grid_search_id: str, experiment_id: str, config_name: str, config: RawTextFile):
        requested_full_path = os.path.realpath(os.path.join(self.top_level_logging_path, str(grid_search_id), str(experiment_id), config_name))

        if FileDataAccess.is_safe_path(base_dir=self.top_level_logging_path, requested_path=requested_full_path):
            os.makedirs(os.path.dirname(requested_full_path), exist_ok=True)
            with open(requested_full_path, "w") as fp:
                fp.writelines(config.content)

        else:
            raise InvalidPathError(f"File path {requested_full_path} is not safe.")

    def get_checkpoint_resource(self, grid_search_id: str, experiment_id: str, epoch: str, checkpoint_resource: CheckpointResource):
        requested_full_path = os.path.realpath(os.path.join(self.top_level_logging_path, str(grid_search_id), str(experiment_id),
                                                            str(epoch), f"{checkpoint_resource}.bin"))
        if FileDataAccess.is_safe_path(base_dir=self.top_level_logging_path, requested_path=requested_full_path):
            if not os.path.isfile(requested_full_path):
                raise InvalidPathError(f"Resource {requested_full_path} not found.")

            generator = FileDataAccess.iterfile(requested_full_path)
            return generator
        else:
            raise InvalidPathError(f"File path {requested_full_path} is not safe.")

    def get_checkpoint_dict_epoch(self, grid_search_id: str, experiment_id: str, epoch: str):
        requested_full_path = os.path.realpath(os.path.join(self.top_level_logging_path, str(grid_search_id), str(experiment_id), str(epoch)))

        if FileDataAccess.is_safe_path(base_dir=self.top_level_logging_path, requested_path=requested_full_path):
            files = FileDataAccess.get_checkpoint_files(requested_full_path, base_path=self.top_level_logging_path)
            return {os.path.basename(file).split(".")[0]: file for file in files}

        else:
            raise InvalidPathError(f"File path {requested_full_path} is not safe.")



from ml_board.backend.messaging.events import Event
from abc import ABC
from typing import List
import os
import json
from pathlib import Path


class EventStorageIF(ABC):

    def add_event(self, event: Event) -> int:
        raise NotImplementedError

    def get_event(self, event_id: int) -> Event:
        raise NotImplementedError

    def iter_generator(self):  # -> Generator[Tuple[str, Event]]:
        raise NotImplementedError

    def length(self) -> int:
        raise NotImplementedError


class ListEventStorage(EventStorageIF):
    def __init__(self):
        self._storage: List[Event] = list()

    def add_event(self, event: Event) -> int:
        # TODO make this thing thread / multiprocessing safe
        # if we use the async_mode eventlet, locking is not necessary
        # see: https://stackoverflow.com/questions/2851499/eventlet-and-locking
        event_id = len(self._storage)
        self._storage.append(event)
        return event_id

    # def get_event(self, event_storage_id: str, event_id: int) -> Event:
    #     if len(self._storage[event_storage_id])-1 < event_id:
    #         raise EventStorageInvalidIndexingError(f"Could not access event at index {event_id}.")
    #     return self._storage[event_storage_id][event_id]

    def iter_generator(self):  # -> Generator[Tuple[str, Event]]:
        current = 0
        while current < len(self._storage):
            yield current, self._storage[current]
            current += 1

    def length(self) -> int:
        return len(self._storage)


class DiscEventStorage(EventStorageIF):
    def __init__(self, parent_dir: str, event_storage_id: str):
        self._event_storage_id = event_storage_id
        # create even_storage log if it does not exist yet
        self.log_dir_path = os.path.join(parent_dir, event_storage_id)
        self.log_path = os.path.join(parent_dir, event_storage_id, "event_storage.log")
        Path(self.log_dir_path).mkdir(parents=True, exist_ok=True)
        Path(self.log_path).touch(exist_ok=True)

    def add_event(self, event: Event) -> int:
        event_id = self.length()  # need to check for possible race condition

        with open(self.log_path, "a") as fd:
            json_event = json.dumps(event)
            fd.write(f"{json_event}\n")

        return event_id

    def iter_generator(self):  # -> Generator[Tuple[str, Event]]:
        event_id = 0
        with open(self.log_path, "r") as fd:
            for line in fd:
                event = json.loads(line)
                yield event_id, event
                event_id += 1

    def length(self) -> int:
        with open(self.log_path, "r") as fd:
            # TODO make this thing thread / multiprocessing safe
            # if we use the async_mode eventlet, locking is not necessary
            # see: https://stackoverflow.com/questions/2851499/eventlet-and-locking
            return len(fd.readlines())  # TODO this is very ineffecient! We should cache this information.

    @property
    def event_storage_ids(self) -> List[str]:
        return self._event_storage_ids


class EventStorageFactory:
    @staticmethod
    def get_list_event_storage() -> EventStorageIF:
        return ListEventStorage()

    @staticmethod
    def get_disc_event_storage(parent_dir: str, event_storage_id: str) -> EventStorageIF:
        return DiscEventStorage(parent_dir, event_storage_id)



from ml_gym.checkpointing.checkpointing import SaveLastEpochOnlyCheckpointingStrategy, SaveAllCheckpointingStrategy, Checkpointing,\
    CheckpointingIF


class CheckpointingStrategyFactory:
    @staticmethod
    def get_save_last_epoch_only_checkpointing_strategy() -> CheckpointingIF:
        strategy = SaveLastEpochOnlyCheckpointingStrategy()
        return Checkpointing(checkpointing_strategy=strategy)

    @staticmethod
    def get_save_all_checkpointing_strategy() -> CheckpointingIF:
        strategy = SaveAllCheckpointingStrategy()
        return Checkpointing(checkpointing_strategy=strategy)



import os
from flask import Flask, request
from flask_socketio import SocketIO, emit, join_room, leave_room, close_room, rooms, disconnect
from ml_board.backend.messaging.event_storage import EventStorageIF, EventStorageFactory
from typing import Any, List, Dict
from engineio.payload import Payload
from pathlib import Path

Payload.max_decode_packets = 10000


class EventSubscriberIF:

    def callback(self):
        raise NotImplementedError


class WebSocketServer:

    def __init__(self, host: str, port: int, async_mode: str, app: Flask, top_level_logging_path: str, cors_allowed_origins: List[str]):
        self._port = port
        self._host = host
        self.app = app
        self._socketio = SocketIO(app,
                                  async_mode=async_mode,
                                  cors_allowed_origins=cors_allowed_origins,
                                  max_http_buffer_size=100000000000)
        self._client_sids = []
        self._top_level_logging_path = top_level_logging_path
        self._room_id_to_event_storage: Dict[str, EventStorageIF] = {
            # "mlgym_event_subscribers": EventStorageFactory.get_disc_event_storage(parent_dir=self.mlgym_event_logging_path)
        }
        self._init_call_backs()

    def emit_server_log_message(self, data):
        emit("server_log_message", data)

    @property
    def client_sids(self) -> List[str]:
        return self._client_sids

    def _send_event_history_to_client(self, client_id: str, room_id: str):
        event_storage = self._room_id_to_event_storage[room_id]
        print(f"=== WEBSOCKET SERVER LOG ===: Sending {event_storage.length()} old messages from room {room_id} to client {client_id}")
        for event_id, event in event_storage.iter_generator():  # TODO make grid search id selectable
            emit('mlgym_event', {'event_id': event_id, 'data': event}, room=client_id)

    def _init_call_backs(self):

        @self._socketio.on("join")
        def on_join(data):
            client_sid = request.sid
            self._client_sids.append(client_sid)
            if 'client_id' in data:
                client_id = data['client_id']
            else:
                client_id = "<unknown>"
            rooms_to_join = data['rooms']
            for room in rooms_to_join:
                if room not in self._room_id_to_event_storage:
                    self._room_id_to_event_storage[room] = EventStorageFactory.get_disc_event_storage(parent_dir=self._top_level_logging_path, event_storage_id=room)
                join_room(room)
            print(f"Client {client_id} joined rooms: {rooms()}")
            print(rooms_to_join)
            self.emit_server_log_message(f"Client {client_id} joined rooms: {rooms()}")
            for room in rooms_to_join:
                self._send_event_history_to_client(client_sid, room)

        @self._socketio.on("leave")
        def on_leave():
            self._client_sids.remove(request.sid)
            # TODO  leave all rooms
            # leave_room(message['room'])
            self.emit_server_log_message("You are now disconnected.")
            disconnect()

        @self._socketio.on("mlgym_event")
        def on_mlgym_event(data):
            grid_search_id = data["payload"]["grid_search_id"]
            if data["event_type"] in set(["experiment_status", "job_status", "experiment_config", "evaluation_result"]):
                print("mlgym_event: " + str(data))
                if grid_search_id not in self._room_id_to_event_storage:
                    self._room_id_to_event_storage[grid_search_id] = EventStorageFactory.get_disc_event_storage(parent_dir=self._top_level_logging_path,
                                                                                                                event_storage_id=grid_search_id)
                event_id = self._room_id_to_event_storage[grid_search_id].add_event(data)
                emit('mlgym_event', {'event_id': event_id, 'data': data}, to=grid_search_id)
            elif data["event_type"] == "checkpoint":
                self.save_checkpoint(checkpoint=data["payload"], path=self._top_level_logging_path)
            else:
                print(f"Unsupported event_type {data['event_type']}")

    # @socketio.event
    # def disconnect_request():
    #     @copy_current_request_context
    #     def can_disconnect():
    #         disconnect()

    #     session['receive_count'] = session.get('receive_count', 0) + 1
    #     # for this emit we use a callback function
    #     # when the callback function is invoked we know that the message has been
    #     # received and it is safe to disconnect
    #     emit('my_response',
    #          {'data': 'Disconnected!', 'count': session['receive_count']},
    #          callback=can_disconnect)-

        @self._socketio.on("ping")
        def on_ping():
            emit('pong')

        @self._socketio.on("client_connected")
        def on_client_connected():
            print(f"Client with SID {request.sid} connnected.")
            self.emit_server_log_message(f"Client with SID {request.sid} connnected.")

        @self._socketio.on("client_disconnected")
        def on_client_disconnected():
            print('Client disconnected', request.sid)
            self._client_sids.remove(request.sid)

    def run(self):
        self._socketio.run(self.app, host=self._host, port=self._port)

    def save_checkpoint(self, checkpoint: Dict[str, Any], path: str):
        grid_search_id = checkpoint["grid_search_id"]
        experiment_id = checkpoint["experiment_id"]
        checkpoint_id = checkpoint["checkpoint_id"]

        full_path = os.path.join(path, grid_search_id, str(experiment_id), str(checkpoint_id))

        for key, stream in checkpoint["checkpoint_streams"].items():
            checkpoint_element_path = os.path.join(full_path, key + ".bin")

            if stream is None:
                if os.path.exists(checkpoint_element_path):
                    os.remove(checkpoint_element_path)
                parent_dir = Path(checkpoint_element_path).parent
                if os.path.exists(parent_dir):
                    if not any(Path(parent_dir).iterdir()):  # if the directory is empty we can also just remove the folder
                        os.rmdir(parent_dir)

            else:  # add new checkpoint or replace checkpoint
                os.makedirs(full_path, exist_ok=True)
                with open(checkpoint_element_path, "wb") as fd:
                    fd.write(stream)


if __name__ == '__main__':
    port = 5000
    async_mode = None
    top_level_logging_path = "event_storage/"
    app = Flask(__name__, template_folder="template")
    app.config['SECRET_KEY'] = 'secret!'

    # thread = socketio.start_background_task(background_thread, )

    ws = WebSocketServer(port=port, async_mode=async_mode, app=app, top_level_logging_path=top_level_logging_path)

    ws.run(app)



from setuptools import find_packages, setup

with open("../README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='mlgym',
    version='0.0.73',

    author='Max Luebbering',
    description="MLgym, a python framework for distributeda and reproducible machine learning model training in research.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=[
        "pytest",
        "pytest-cov",
        "torch",
        "h5py",
        "tqdm",
        "pyyaml",
        "datastack",
        "scipy",
        "scikit-learn",
        "python-socketio",
        "fastapi",
        "uvicorn",
        "requests",
        "python-multipart",
        "flask",
        "Flask-SocketIO",
        "eventlet",
        "python-socketio[client]",
        "gunicorn"
    ],
    python_requires=">=3.7",
    include_package_data=True,
    scripts=['ml_board/backend/starter_scripts/ml_board', 'ml_board/backend/starter_scripts/ml_board_ws_endpoint', 'ml_board/backend/starter_scripts/ml_board_rest_endpoint'],
    url="https://github.com/mlgym/mlgym"
)



from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Dict, List
import requests
from http import HTTPStatus
from ml_gym.error_handling.exception import NetworkError, DataIntegrityError
from abc import ABC
from ml_board.backend.restful_api.data_models import RawTextFile, FileFormat, ExperimentStatus, CheckpointResource


class GridSearchAPIClientIF(ABC):

    def get_config(self, grid_search_id: str, config_name: str):
        raise NotImplementedError

    def add_config_string(self, grid_search_id: str, config_name: str, config: Dict, experiment_id: int = None) -> Dict:
        raise NotImplementedError

    def get_validation_config(self, grid_search_id: str):
        raise NotImplementedError

    def get_checkpoint_resource(self, grid_search_id: str, experiment_id: str,  checkpoint_id: int,
                                checkpoint_resource: CheckpointResource):
        raise NotImplementedError

    def get_full_checkpoint(self, grid_search_id: str, experiment_id: str,  checkpoint_id: int):
        raise NotImplementedError

    def get_unfinished_experiments(self, grid_search_id: str):
        raise NotImplementedError

    def get_experiment_statuses(self, grid_search_id: str) -> List[ExperimentStatus]:
        raise NotImplementedError


class GridSearchRestfulAPIClient(GridSearchAPIClientIF):

    def __init__(self, endpoint: str):
        self.endpoint = endpoint

    def _get_json_resource(url) -> Dict:
        response = requests.get(url)
        if response.status_code != HTTPStatus.OK:
            raise NetworkError(f"Server responded with error code {response.status_code}")
        else:
            try:
                grid_search_config = response.json()
            except ValueError as e:
                raise DataIntegrityError(f"Could not cast response from {url} to json.") from e
            return grid_search_config

    def _get_binary_resource(url) -> bytes:
        response = requests.get(url)
        if response.status_code != HTTPStatus.OK:
            raise NetworkError(f"Server responded with error code {response.status_code}")
        else:
            return response.content

    def _set_raw_text_file_resource(url: str, payload: Dict) -> Dict:
        response = requests.put(url=url, json=payload)
        if response.status_code != HTTPStatus.OK:
            raise NetworkError(f"Server responded with error code {response.status_code}")

    def get_config(self, grid_search_id: str, config_name: str) -> str:
        url = f"{self.endpoint}/{grid_search_id}/{config_name}"
        return GridSearchRestfulAPIClient._get_json_resource(url)

    # def add_grid_search_config(self, grid_search_id: str, grid_search_config: Dict) -> Dict:
    #     url = f"{self.endpoint}/grid_searches/{grid_search_id}/gs_config"
    #     return GridSearchRestfulAPIClient._set_raw_text_file_resource(url, grid_search_config)

    def add_config_string(self, grid_search_id: str, config_name: str, config: str, file_format: FileFormat,
                          experiment_id: int = None) -> Dict:
        payload_dict = RawTextFile(file_format=file_format, content=config).dict()
        if experiment_id is None:
            url = f"{self.endpoint}/grid_searches/{grid_search_id}/{config_name}"
        else:
            url = f"{self.endpoint}/grid_searches/{grid_search_id}/{experiment_id}/{config_name}"

        return GridSearchRestfulAPIClient._set_raw_text_file_resource(url, payload_dict)

    def get_validation_config(self, grid_search_id: str):
        url = f"{self.endpoint}/{grid_search_id}/validation_config"
        return GridSearchRestfulAPIClient._get_json_resource(url)

    def get_experiments(self, grid_search_id: str):
        url = f"{self.endpoint}/{grid_search_id}/experiments"
        return GridSearchRestfulAPIClient._get_json_resource(url)

    def get_full_checkpoint(self, grid_search_id: str, experiment_id: str,  checkpoint_id: int):
        url = f"{self.endpoint}/checkpoints/{grid_search_id}/{experiment_id}/{checkpoint_id}"
        return GridSearchRestfulAPIClient._get_json_resource(url)

    def get_checkpoint_resource(self, grid_search_id: str, experiment_id: str,  checkpoint_id: int,
                                checkpoint_resource: CheckpointResource):
        url = f"{self.endpoint}/checkpoints/{grid_search_id}/{experiment_id}/{checkpoint_id}/{checkpoint_resource}"
        return GridSearchRestfulAPIClient._get_binary_resource(url)

    def get_experiment_statuses(self, grid_search_id: str) -> List[ExperimentStatus]:
        url = f"{self.endpoint}/grid_searches/{grid_search_id}/experiments"
        response = GridSearchRestfulAPIClient._get_json_resource(url)
        experiment_statuses = [ExperimentStatus(**r) for r in response]
        return experiment_statuses


class GridSearchAPIClientConstructableIF(ABC):

    @abstractmethod
    def construct(self) -> GridSearchAPIClientIF:
        raise NotImplementedError


class GridSearchAPIClientType(Enum):
    GRID_SEARCH_RESTFUL_API_CLIENT = GridSearchRestfulAPIClient


@dataclass
class GridSearchAPIClientConfig:
    api_client_type: GridSearchAPIClientType
    api_client_config: Dict


@dataclass
class GridSearchAPIClientConstructable(GridSearchAPIClientConstructableIF):
    config: GridSearchAPIClientConfig

    def construct(self) -> GridSearchAPIClientIF:
        return self.config.api_client_type.value(**self.config.api_client_config)
